@@ -2,6 +2,6 @@ import os
 from pathlib import Path
 
 # charles
-DEFAULT_ROOT_PATH = Path(os.path.expanduser(os.getenv("CHIA_ROOT", "~/.kik/mainnet"))).resolve()
+DEFAULT_ROOT_PATH = Path(os.path.expanduser(os.getenv("CHIA_ROOT", "~/.kiwi/mainnet"))).resolve()
 
 DEFAULT_KEYS_ROOT_PATH = Path(os.path.expanduser(os.getenv("CHIA_KEYS_ROOT", "~/.kiwi_keys"))).resolve()
