@@ -4,10 +4,12 @@ import logging
 import socket
 import time
 import traceback
+from decimal import Decimal
 from pathlib import Path
 from typing import Callable, Dict, List, Optional, Set, Tuple, Union
 
-from blspy import PrivateKey
+from blspy import G1Element, PrivateKey
+
 from chia.consensus.block_record import BlockRecord
 from chia.consensus.blockchain_interface import BlockchainInterface
 from chia.consensus.constants import ConsensusConstants
@@ -21,7 +23,7 @@ from chia.daemon.keychain_proxy import (
     wrap_local_keychain,
 )
 from chia.pools.pool_puzzles import SINGLETON_LAUNCHER_HASH
-from chia.protocols import wallet_protocol
+from chia.protocols import farmer_protocol, wallet_protocol
 from chia.protocols.full_node_protocol import RequestProofOfWeight, RespondProofOfWeight
 from chia.protocols.protocol_message_types import ProtocolMessageTypes
 from chia.protocols.wallet_protocol import (
@@ -47,11 +49,12 @@ from chia.types.peer_info import PeerInfo
 from chia.util.byte_types import hexstr_to_bytes
 from chia.util.check_fork_next_block import check_fork_next_block
 from chia.util.errors import Err, ValidationError
-from chia.util.ints import uint32, uint128
+from chia.util.ints import uint32, uint64, uint128
 from chia.util.keychain import Keychain
 from chia.util.lru_cache import LRUCache
 from chia.util.merkle_set import MerkleSet, confirm_included_already_hashed, confirm_not_included_already_hashed
 from chia.util.path import mkdir, path_from_root
+from chia.util.profiler import profile_task
 from chia.wallet.block_record import HeaderBlockRecord
 from chia.wallet.derivation_record import DerivationRecord
 from chia.wallet.settings.settings_objects import BackupInitialized
@@ -61,7 +64,6 @@ from chia.wallet.util.wallet_types import WalletType
 from chia.wallet.wallet_action import WalletAction
 from chia.wallet.wallet_blockchain import ReceiveBlockResult
 from chia.wallet.wallet_state_manager import WalletStateManager
-from chia.util.profiler import profile_task
 
 
 class WalletNode:
@@ -236,7 +238,10 @@ class WalletNode:
         self.peer_task = asyncio.create_task(self._periodically_check_full_node())
         self.sync_event = asyncio.Event()
         self.sync_task = asyncio.create_task(self.sync_job())
-        self.logged_in_fingerprint = fingerprint
+        if fingerprint is None:
+            self.logged_in_fingerprint = private_key.get_g1().get_fingerprint()
+        else:
+            self.logged_in_fingerprint = fingerprint
         self.logged_in = True
         return True
 
@@ -395,7 +400,12 @@ class WalletNode:
                 self.config["full_node_peer"]["port"],
             )
             peers = [c.get_peer_info() for c in self.server.get_full_node_connections()]
-            full_node_resolved = PeerInfo(socket.gethostbyname(full_node_peer.host), full_node_peer.port)
+            # If full_node_peer is already an address, use it, otherwise
+            # resolve it here.
+            if full_node_peer.is_valid():
+                full_node_resolved = full_node_peer
+            else:
+                full_node_resolved = PeerInfo(socket.gethostbyname(full_node_peer.host), full_node_peer.port)
             if full_node_peer in peers or full_node_resolved in peers:
                 self.log.info(f"Will not attempt to connect to other nodes, already connected to {full_node_peer}")
                 for connection in self.server.get_full_node_connections():
@@ -408,6 +418,33 @@ class WalletNode:
                 return True
         return False
 
+    async def update_stakings(self, peer: WSChiaConnection, height: uint64, farmer_public_key: G1Element) -> None:
+        "fetch staking"
+        height = height - 1 if height > 0 else 0
+        blockchain = self.wallet_state_manager.blockchain
+
+        # calculate blocks from cache
+        blocks = 0
+        if blockchain.get_peak_height() is not None:
+            block_range = self.constants.STAKING_ESTIMATE_BLOCK_RANGE
+            curr: Optional[BlockRecord] = blockchain.try_block_record(blockchain.height_to_hash(height))
+            begin_height = max((curr.height if curr is not None else 0) - block_range, 1)
+            while curr is not None and curr.height > begin_height:
+                if curr.farmer_public_key == farmer_public_key:
+                    blocks += 1
+                curr = blockchain.try_block_record(curr.prev_hash)
+
+        res: Optional[farmer_protocol.FarmerStakings] = await peer.request_stakings(
+            farmer_protocol.RequestStakings(
+                public_keys=[farmer_public_key],
+                height=height,
+                blocks=blocks,
+            )
+        )
+        if res is None or not isinstance(res, farmer_protocol.FarmerStakings):
+            raise ValueError("Peer returned no response")
+        self.wallet_state_manager.blockchain.stakings.update({bytes(k): Decimal(v) for k, v in res.stakings})
+
     async def complete_blocks(self, header_blocks: List[HeaderBlock], peer: WSChiaConnection):
         if self.wallet_state_manager is None:
             return None
@@ -416,6 +453,10 @@ class WalletNode:
         trusted = self.server.is_trusted_peer(peer, self.config["trusted_peers"])
         async with self.wallet_state_manager.blockchain.lock:
             for block in header_blocks:
+                await self.update_stakings(
+                    peer, block.height, block.reward_chain_block.proof_of_space.farmer_public_key
+                )
+
                 if block.is_transaction_block:
                     # Find additions and removals
                     (additions, removals,) = await self.wallet_state_manager.get_filter_additions_removals(
@@ -702,6 +743,11 @@ class WalletNode:
 
         for i in range(len(header_blocks)):
             header_block = header_blocks[i]
+
+            await self.update_stakings(
+                peer, header_block.height, header_block.reward_chain_block.proof_of_space.farmer_public_key
+            )
+
             if not trusted and pre_validation_results is not None and pre_validation_results[i].error is not None:
                 raise ValidationError(Err(pre_validation_results[i].error))
 
@@ -757,7 +803,7 @@ class WalletNode:
                 advanced_peak = True
                 self.wallet_state_manager.state_changed("new_block")
             elif result == ReceiveBlockResult.INVALID_BLOCK:
-                raise ValueError("Value error peer sent us invalid block")
+                raise ValueError(f"Value error peer sent us invalid block {error} {fork_h}")
         if advanced_peak:
             await self.wallet_state_manager.create_more_puzzle_hashes()
         return True, advanced_peak
@@ -886,10 +932,14 @@ class WalletNode:
             all_added_coins = await self.get_additions(peer, block, [], get_all_additions=True)
             assert all_added_coins is not None
             if all_added_coins is not None:
-
+                all_added_coin_parents = [c.parent_coin_info for c in all_added_coins]
                 for coin in all_added_coins:
                     # This searches specifically for a launcher being created, and adds the solution of the launcher
-                    if coin.puzzle_hash == SINGLETON_LAUNCHER_HASH and coin.parent_coin_info in removed_coin_ids:
+                    if (
+                        coin.puzzle_hash == SINGLETON_LAUNCHER_HASH  # Check that it's a launcher
+                        and coin.name() in all_added_coin_parents  # Check that it's ephemermal
+                        and coin.parent_coin_info in removed_coin_ids  # Check that an interesting coin created it
+                    ):
                         cs: CoinSpend = await self.fetch_puzzle_solution(peer, block.height, coin)
                         additional_coin_spends.append(cs)
                         # Apply this coin solution, which might add things to interested list
@@ -963,7 +1013,7 @@ class WalletNode:
         for coin in additions:
             puzzle_store = self.wallet_state_manager.puzzle_store
             record_info: Optional[DerivationRecord] = await puzzle_store.get_derivation_record_for_puzzle_hash(
-                coin.puzzle_hash.hex()
+                coin.puzzle_hash
             )
             if record_info is not None and record_info.wallet_type == WalletType.COLOURED_COIN:
                 # TODO why ?
